@@ -1,11 +1,9 @@
-import os
 import re
-import json
 import hashlib
 from datetime import datetime, timedelta
 from enum import Enum as EnumBaseClass
 from functools import wraps
-from typing import Optional, List, Union, Callable
+from typing import Dict, Optional, List, Union, Callable
 
 from flask import current_app as app, Flask
 from google.cloud.storage import Blob
@@ -26,9 +24,10 @@ from sqlalchemy import (
     asc,
     desc,
     update,
-    or_,
     case,
+    select,
     literal_column,
+    not_,
 )
 from sqlalchemy.exc import IntegrityError
 from sqlalchemy.ext.hybrid import hybrid_property
@@ -39,11 +38,19 @@ from sqlalchemy.orm.session import Session
 from sqlalchemy.orm.query import Query
 from sqlalchemy.sql import expression, text
 from sqlalchemy.dialects.postgresql import JSONB, UUID
+from sqlalchemy.engine import ResultProxy
 from sqlalchemy.engine.interfaces import ExecutionContext
 
 from cidc_schemas import prism, unprism, json_validation
+from sqlalchemy.sql.elements import literal
 
-from .facets import get_facet_groups_for_paths, facet_groups_to_names
+from .files import (
+    get_facet_groups_for_paths,
+    facet_groups_to_names,
+    details_dict,
+    FilePurpose,
+    FACET_NAME_DELIM,
+)
 from ..config.db import BaseModel
 from ..config.settings import (
     PAGINATION_PAGE_SIZE,
@@ -146,24 +153,28 @@ class CommonColumns(BaseModel):  # type: ignore
 
     @classmethod
     @with_default_session
-    def list(
+    def list(cls, session: Session, **pagination_args):
+        """List records in this table, with pagination support."""
+        query = session.query(cls)
+        query = cls._add_pagination_filters(query, **pagination_args)
+        return query.all()
+
+    @classmethod
+    def _add_pagination_filters(
         cls,
-        session: Session,
+        query: Query,
         page_num: int = 0,
         page_size: int = PAGINATION_PAGE_SIZE,
         sort_field: Optional[str] = None,
         sort_direction: Optional[str] = None,
         filter_: Callable[[Query], Query] = lambda q: q,
-    ):
-        """List records in this table, with pagination support."""
+    ) -> Query:
         # Enforce positive page numbers
         page_num = 0 if page_num < 0 else page_num
 
         # Enforce maximum page size
         page_size = min(page_size, MAX_PAGINATION_PAGE_SIZE)
 
-        query = session.query(cls)
-
         # Handle sorting
         if sort_field:
             # Get the attribute from the class, in case this is a hybrid attribute
@@ -180,7 +191,7 @@ class CommonColumns(BaseModel):  # type: ignore
         query = query.offset(page_num * page_size)
         query = query.limit(page_size)
 
-        return query.all()
+        return query
 
     @classmethod
     @with_default_session
@@ -263,8 +274,8 @@ class Users(CommonColumns):
     @with_default_session
     def find_by_email(email: str, session: Session) -> Optional:
         """
-            Search for a record in the Users table with the given email.
-            If found, return the record. If not found, return None.
+        Search for a record in the Users table with the given email.
+        If found, return the record. If not found, return None.
         """
         user = session.query(Users).filter_by(email=email).first()
         return user
@@ -273,9 +284,9 @@ class Users(CommonColumns):
     @with_default_session
     def create(profile: dict, session: Session):
         """
-            Create a new record for a user if one doesn't exist
-            for the given email. Return the user record associated
-            with that email.
+        Create a new record for a user if one doesn't exist
+        for the given email. Return the user record associated
+        with that email.
         """
         email = profile.get("email")
         first_n = profile.get("given_name")
@@ -360,6 +371,22 @@ class Permissions(CommonColumns):
                 orig=f"`granted_to_user` user must exist, but no user found with id {self.granted_to_user}",
             )
 
+        grantor = None
+        if self.granted_by_user is not None:
+            grantor = Users.find_by_id(self.granted_by_user)
+        else:
+            raise IntegrityError(
+                params=None,
+                statement=None,
+                orig=f"`granted_by_user` user must be given",
+            )
+        if grantor is None:
+            raise IntegrityError(
+                params=None,
+                statement=None,
+                orig=f"`granted_by_user` user must exist, but no user found with id {self.granted_by_user}",
+            )
+
         # A user can only have 20 granular permissions at a time, due to GCS constraints
         if (
             len(Permissions.find_for_user(self.granted_to_user))
@@ -371,6 +398,10 @@ class Permissions(CommonColumns):
                 orig=f"{grantee.email} has greater than or equal to the maximum number of allowed granular permissions ({GOOGLE_MAX_DOWNLOAD_PERMISSIONS}). Remove unused permissions to add others.",
             )
 
+        print(
+            f"admin-action: {grantor.email} gave {grantee.email} the permission {self.upload_type} on {self.trial_id}"
+        )
+
         # Always commit, because we don't want to grant IAM download unless this insert succeeds.
         super().insert(session=session, commit=True, compute_etag=compute_etag)
 
@@ -383,7 +414,9 @@ class Permissions(CommonColumns):
             raise IAMException("IAM grant failed.") from e
 
     @with_default_session
-    def delete(self, session: Session, commit: bool = True):
+    def delete(
+        self, deleted_by: Union[Users, int], session: Session, commit: bool = True
+    ):
         """
         Delete this permission record from the database and revoke the corresponding IAM policy binding
         on the GCS data bucket.
@@ -394,6 +427,13 @@ class Permissions(CommonColumns):
         if grantee is None:
             raise NoResultFound(f"no user with id {self.granted_to_user}")
 
+        if not isinstance(deleted_by, Users):
+            deleted_by_user = Users.find_by_id(deleted_by)
+        else:
+            deleted_by_user = deleted_by
+        if deleted_by_user is None:
+            raise NoResultFound(f"no user with id {deleted_by}")
+
         try:
             # Revoke IAM permission in GCS
             revoke_download_access(grantee.email, self.trial_id, self.upload_type)
@@ -402,6 +442,9 @@ class Permissions(CommonColumns):
                 "IAM revoke failed, and permission db record not removed."
             ) from e
 
+        print(
+            f"admin-action: {deleted_by_user.email} removed from {grantee.email} the permission {self.upload_type} on {self.trial_id}"
+        )
         super().delete(session=session, commit=True)
 
     @staticmethod
@@ -452,10 +495,14 @@ class ValidationMultiError(Exception):
     pass
 
 
-trial_metadata_validator: json_validation._Validator = json_validation.load_and_validate_schema(
-    "clinical_trial.json", return_validator=True
+trial_metadata_validator: json_validation._Validator = (
+    json_validation.load_and_validate_schema(
+        "clinical_trial.json", return_validator=True
+    )
 )
 
+FileBundle = Dict[str, Dict[FilePurpose, List[int]]]
+
 
 class TrialMetadata(CommonColumns):
     __tablename__ = "trial_metadata"
@@ -478,7 +525,7 @@ class TrialMetadata(CommonColumns):
     @with_default_session
     def find_by_trial_id(trial_id: str, session: Session):
         """
-            Find a trial by its CIMAC id.
+        Find a trial by its CIMAC id.
         """
         return session.query(TrialMetadata).filter_by(trial_id=trial_id).first()
 
@@ -486,7 +533,7 @@ class TrialMetadata(CommonColumns):
     @with_default_session
     def select_for_update_by_trial_id(trial_id: str, session: Session):
         """
-            Find a trial by its CIMAC id.
+        Find a trial by its CIMAC id.
         """
         try:
             trial = (
@@ -505,9 +552,9 @@ class TrialMetadata(CommonColumns):
         trial_id: str, assay_patch: dict, session: Session, commit: bool = False
     ):
         """
-            Applies assay updates to the metadata object from the trial with id `trial_id`.
+        Applies assay updates to the metadata object from the trial with id `trial_id`.
 
-            TODO: apply this update directly to the not-yet-existent TrialMetadata.manifest field
+        TODO: apply this update directly to the not-yet-existent TrialMetadata.manifest field
         """
         return TrialMetadata._patch_trial_metadata(
             trial_id, assay_patch, session=session, commit=commit
@@ -519,9 +566,9 @@ class TrialMetadata(CommonColumns):
         trial_id: str, manifest_patch: dict, session: Session, commit: bool = False
     ):
         """
-            Applies manifest updates to the metadata object from the trial with id `trial_id`.
+        Applies manifest updates to the metadata object from the trial with id `trial_id`.
 
-            TODO: apply this update directly to the not-yet-existent TrialMetadata.assays field
+        TODO: apply this update directly to the not-yet-existent TrialMetadata.assays field
         """
         return TrialMetadata._patch_trial_metadata(
             trial_id, manifest_patch, session=session, commit=commit
@@ -533,11 +580,11 @@ class TrialMetadata(CommonColumns):
         trial_id: str, json_patch: dict, session: Session, commit: bool = False
     ):
         """
-            Applies updates to the metadata object from the trial with id `trial_id`
-            and commits current session.
+        Applies updates to the metadata object from the trial with id `trial_id`
+        and commits current session.
 
-            TODO: remove this function and dependency on it, in favor of separate assay
-            and manifest patch strategies.
+        TODO: remove this function and dependency on it, in favor of separate assay
+        and manifest patch strategies.
         """
 
         trial = TrialMetadata.select_for_update_by_trial_id(trial_id, session=session)
@@ -564,7 +611,7 @@ class TrialMetadata(CommonColumns):
         trial_id: str, metadata_json: dict, session: Session, commit: bool = True
     ):
         """
-            Create a new clinical trial metadata record.
+        Create a new clinical trial metadata record.
         """
 
         print(f"Creating new trial metadata with id {trial_id}")
@@ -606,6 +653,45 @@ class TrialMetadata(CommonColumns):
             raise NoResultFound(f"No trial found with id {trial_id}")
         return unprism.unprism_samples(trial.metadata_json)
 
+    file_bundle: Optional[FileBundle]
+
+    @classmethod
+    @with_default_session
+    def list_with_file_bundles(cls, session: Session, **pagination_args):
+        """List `TrialMetadata` records from the database along with their `file_bundle`s."""
+        # Build a query that collects trials along with all files for that trial
+        file_bundle_query = DownloadableFiles.build_file_bundle_query()
+        query = session.query(cls, file_bundle_query.c.file_bundle).join(
+            file_bundle_query, TrialMetadata.trial_id == file_bundle_query.c.trial_id
+        )
+
+        # Apply pagination options to the query
+        query = cls._add_pagination_filters(query, **pagination_args)
+
+        # Run the query
+        query_results = query.all()
+
+        # Add file bundles to the `file_bundle` attribute on each trial record
+        trials: List[TrialMetadata] = []
+        for trial, file_bundle in query_results:
+            trial.file_bundle = file_bundle
+            trials.append(trial)
+
+        return trials
+
+    @classmethod
+    def build_trial_filter(cls, user: Users, trial_ids: List[str] = []):
+        filters = []
+        if trial_ids:
+            filters.append(cls.trial_id.in_(trial_ids))
+        if not user.is_admin() and not user.is_nci_user():
+            permitted_trials = select([Permissions.trial_id]).where(
+                Permissions.granted_to_user == user.id
+            )
+            filters.append(cls.trial_id.in_(permitted_trials))
+        # possible TODO: filter by assays in a trial
+        return lambda q: q.filter(*filters)
+
 
 class UploadJobStatus(EnumBaseClass):
     STARTED = "started"
@@ -772,15 +858,34 @@ class UploadJobs(CommonColumns):
 
     @staticmethod
     @with_default_session
-    def merge_extra_metadata(job_id, files, session):
+    def merge_extra_metadata(job_id: int, files: dict, session: Session):
+        """
+        Args:
+            job_id: the ID of the UploadJob to merge
+            files: mapping from uuid of the artifact-to-update to metadata file-to-update-from
+            session: the current session; uses default if not passed
+        Returns:
+            None
+        Raises:
+            ValueError
+                if `job_id` doesn't exist or is already merged
+                from prism.merge_artifact_extra_metadata
+        """
 
         job = UploadJobs.find_by_id(job_id, session=session)
 
+        if job is None or job.status == UploadJobStatus.MERGE_COMPLETED:
+            raise ValueError(f"Upload job {job_id} doesn't exist or is already merged")
+
         print(f"About to merge extra md to {job.id}/{job.status}")
 
         for uuid, file in files.items():
             print(f"About to parse/merge extra md on {uuid}")
-            job.metadata_patch, updated_artifact, _ = prism.merge_artifact_extra_metadata(
+            (
+                job.metadata_patch,
+                updated_artifact,
+                _,
+            ) = prism.merge_artifact_extra_metadata(
                 job.metadata_patch, uuid, job.upload_type, file
             )
             print(f"Updated md for {uuid}: {updated_artifact.keys()}")
@@ -824,7 +929,7 @@ class UploadJobs(CommonColumns):
 
 class DownloadableFiles(CommonColumns):
     """
-    Store required fields from: 
+    Store required fields from:
     https://github.com/CIMAC-CIDC/cidc-schemas/blob/master/cidc_schemas/schemas/artifacts/artifact_core.json
     """
 
@@ -843,7 +948,7 @@ class DownloadableFiles(CommonColumns):
     facet_group = Column(String, nullable=False)
     # NOTE: this column actually has type CITEXT.
     data_format = Column(String, nullable=False)
-    additional_metadata = Column(JSONB, nullable=True)
+    additional_metadata = Column(JSONB, nullable=False)
     # TODO rename upload_type, because we store manifests in there too.
     # NOTE: this column actually has type CITEXT.
     upload_type = Column(String, nullable=False)
@@ -879,9 +984,82 @@ class DownloadableFiles(CommonColumns):
     def data_category(cls):
         return DATA_CATEGORY_CASE_CLAUSE
 
+    @hybrid_property
+    def data_category_prefix(self):
+        """
+        The overarching data category for a file. E.g., files with `upload_type` of
+        "cytof"` and `"cytof_analyis"` should both have a `data_category_prefix` of `"CyTOF"`.
+        """
+        if self.data_category is None:
+            return None
+        return self.data_category.split(FACET_NAME_DELIM, 1)[0]
+
+    @data_category_prefix.expression
+    def data_category_prefix(cls):
+        return func.split_part(DATA_CATEGORY_CASE_CLAUSE, FACET_NAME_DELIM, 1)
+
+    @hybrid_property
+    def file_purpose(self):
+        return details_dict.get(self.facet_group).file_purpose
+
+    @file_purpose.expression
+    def file_purpose(cls):
+        return FILE_PURPOSE_CASE_CLAUSE
+
     @property
-    def flat_object_url(self):
-        return self.object_url.replace("/", "_")
+    def cimac_id(self):
+        """
+        Extract the `cimac_id` associated with this file, if any, by searching the file's 
+        additional metadata for a field with a key like `<some>.<path>.cimac_id`.
+
+        NOTE: this is not a sqlalchemy hybrid_property, and it can't be used directly in queries.
+        """
+        for key, value in self.additional_metadata.items():
+            if key.endswith("cimac_id"):
+                return value
+        return None
+
+    @with_default_session
+    def get_related_files(self, session: Session) -> list:
+        """
+        Return a list of file records related to this file. We could define "related"
+        in any number of ways, but currently, a related file:
+            * is sample-specific, and relates to the same sample as this file if this file 
+              has an associated `cimac_id`.
+            * isn't sample-specific, and relates to the same `data_category_prefix`.
+        """
+        # If this file has an associated sample, get other files associated with that sample.
+        # Otherwise, get other non-sample-specific files for this trial and data category.
+        if self.cimac_id is not None:
+            query = text(
+                "SELECT DISTINCT downloadable_files.* "
+                "FROM downloadable_files, LATERAL jsonb_each_text(additional_metadata) addm_kv "
+                "WHERE addm_kv.value LIKE :cimac_id AND trial_id = :trial_id AND id != :id"
+            )
+            params = {
+                "cimac_id": f"%{self.cimac_id}",
+                "trial_id": self.trial_id,
+                "id": self.id,
+            }
+            related_files = result_proxy_to_models(
+                session.execute(query, params), DownloadableFiles
+            )
+        else:
+            not_sample_specific = not_(
+                literal_column("additional_metadata::text").like('%.cimac_id":%')
+            )
+            related_files = (
+                session.query(DownloadableFiles)
+                .filter(
+                    DownloadableFiles.trial_id == self.trial_id,
+                    DownloadableFiles.data_category_prefix == self.data_category_prefix,
+                    DownloadableFiles.id != self.id,
+                    not_sample_specific,
+                )
+                .all()
+            )
+
+        return related_files
 
     @staticmethod
     def build_file_filter(
@@ -891,12 +1069,12 @@ class DownloadableFiles(CommonColumns):
         Build a file filter function based on the provided parameters. The resultant
         filter can then be passed as the `filter_` argument of `DownloadableFiles.list`
         or `DownloadableFiles.count`.
-        
+
         Args:
             trial_ids: if provided, the filter will include only files with these trial IDs.
             upload_types: if provided, the filter will include only files with these upload types.
             analysis_friendly: if True, the filter will include only files that are "analysis-friendly".
-            non_admin_user_id: if provided, the filter will include only files that satisfy 
+            non_admin_user_id: if provided, the filter will include only files that satisfy
                 this user's data access permissions.
         Returns:
             A function that adds filters to a query against the DownloadableFiles table.
@@ -938,11 +1116,10 @@ class DownloadableFiles(CommonColumns):
 
         # Filter out keys that aren't columns
         supported_columns = DownloadableFiles.__table__.columns.keys()
-        filtered_metadata = {
-            "trial_id": trial_id,
-            "upload_type": upload_type,
-            "additional_metadata": additional_metadata,
-        }
+        filtered_metadata = {"trial_id": trial_id, "upload_type": upload_type}
+        if additional_metadata not in ("null", None, {}):
+            filtered_metadata["additional_metadata"] = additional_metadata
+
         for key, value in file_metadata.items():
             if key in supported_columns:
                 filtered_metadata[key] = value
@@ -1020,14 +1197,106 @@ class DownloadableFiles(CommonColumns):
     @with_default_session
     def get_by_object_url(object_url: str, session: Session):
         """
-        Look up the downloadable file record associated with 
+        Look up the downloadable file record associated with
         the given GCS object url.
         """
         return session.query(DownloadableFiles).filter_by(object_url=object_url).one()
 
+    @classmethod
+    @with_default_session
+    def list_object_urls(
+        cls, ids: List[int], session: Session, filter_: Callable[[Query], Query]
+    ) -> List[str]:
+        """Get all object_urls for a batch of downloadable file record IDs"""
+        query = session.query(cls.object_url).filter(cls.id.in_(ids))
+        query = filter_(query)
+        return [r[0] for r in query.all()]
+
+    @classmethod
+    def build_file_bundle_query(cls) -> Query:
+        """
+        Build a query that selects nested file bundles from the downloadable files table.
+        The `file_bundles` query below should produce one bundle per unique `trial_id` that
+        appears in the downloadable files table. Each bundle will have shape like:
+        ```
+          {
+              <type 1>: {
+                <purpose 1>: [<file id 1>, <file id 2>, ...],
+                <purpose 2>: [...]
+              },
+              <type 2>: {...}
+          }
+        ```
+        where "type" is something like `"Olink"` or `"Participants Info"` and "purpose" is a `FilePurpose` string.
+        """
+        tid_col, type_col, purp_col, ids_col, purps_col = (
+            literal_column("trial_id"),
+            literal_column("type"),
+            literal_column("purpose"),
+            literal_column("ids"),
+            literal_column("purposes"),
+        )
+
+        id_bundles = (
+            select(
+                [
+                    cls.trial_id,
+                    cls.data_category_prefix.label(type_col.key),
+                    cls.file_purpose.label(purp_col.key),
+                    func.json_agg(cls.id).label(ids_col.key),
+                ]
+            )
+            .group_by(cls.trial_id, cls.data_category_prefix, cls.file_purpose)
+            .alias("id_bundles")
+        )
+        purpose_bundles = (
+            select(
+                [
+                    tid_col,
+                    type_col,
+                    func.json_object_agg(
+                        func.coalesce(purp_col, "miscellaneous"), ids_col
+                    ).label(purps_col.key),
+                ]
+            )
+            .select_from(id_bundles)
+            .group_by(tid_col, type_col)
+            .alias("purpose_bundles")
+        )
+        file_bundles = (
+            select(
+                [
+                    tid_col.label(tid_col.key),
+                    func.json_object_agg(
+                        func.coalesce(type_col, "other"), purps_col
+                    ).label("file_bundle"),
+                ]
+            )
+            .select_from(purpose_bundles)
+            .group_by(tid_col)
+            .alias("file_bundles")
+        )
+        return file_bundles
+
 
 # Query clause for computing a downloadable file's data category.
 # Used above in the DownloadableFiles.data_category computed property.
 DATA_CATEGORY_CASE_CLAUSE = case(
     [(DownloadableFiles.facet_group == k, v) for k, v in facet_groups_to_names.items()]
 )
+
+# Query clause for computing a downloadable file's file purpose.
+# Used above in the DownloadableFiles.file_purpose computed property.
+FILE_PURPOSE_CASE_CLAUSE = case(
+    [
+        (DownloadableFiles.facet_group == facet_group, file_details.file_purpose)
+        for facet_group, file_details in details_dict.items()
+    ]
+)
+
+
+def result_proxy_to_models(
+    result_proxy: ResultProxy, model: BaseModel
+) -> List[BaseModel]:
+    """Materialize a sqlalchemy `result_proxy` iterable as a list of `model` instances"""
+    return [model(**dict(row_proxy.items())) for row_proxy in result_proxy]
